#!/usr/bin/env python

import sys, os, pickle, popen2

PICKLE_DIR = './pkls/'  # Intermediate result folder
COLS       = 1000;       # Split size (number of data set columns)
NPROCS     = 2;         # Number of local processes to launch concurrently

### compare function used in list.sort(): sort by element[1]
def value_cmpf(x,y):
	if x[1]>y[1]: return -1
	if x[1]<y[1]: return 1
	return 0

def sortFscores( score_dict ):
	score_tuples = score_dict.items()
	score_tuples.sort(value_cmpf)
	feat_v = score_tuples
	for i in range(len(feat_v)): feat_v[i]=score_tuples[i][0]
	return score_tuples, feat_v

### cal importance of features
### return fscore_dict and feat with desc order
def cal_feat_imp(label,sample):
	score_dict=cal_Fscore(label,sample)
	score_tuples, feat_v = sortFscores( score_dict )
	return score_dict, feat_v

### return a dict containing F_j
def cal_Fscore(labels,samples):
	data_num=float(len(samples))
	p_num = {} #key: label;  value: data num
	sum_f = [] #index: feat_idx;  value: sum
	sum_l_f = {} #dict of lists.  key1: label; index2: feat_idx; value: sum
	sumq_l_f = {} #dict of lists.  key1: label; index2: feat_idx; value: sum of square
	F={} #key: feat_idx;  valud: fscore
	max_idx = -1
	min_idx = 10E9 # something large 

	### pass 1: count number of exemplars in each class and find max index of features
	for p in range(len(samples)): # for every exemplar
		label = labels[p]; point = samples[p]
		if p_num.has_key(label): 
			p_num[label] += 1
		else: 
			p_num[label] = 1

		for f in point.keys(): # for every feature
			if f>max_idx: max_idx=f
			if f<min_idx: min_idx=f
	
	### now p_num, min_idx and max_idx are set
	### initialize variables
	sum_f = [0 for i in range(max_idx)]
	for la in p_num.keys():
		sum_l_f[la] = [0 for i in range(max_idx)] # init some zero lists
		sumq_l_f[la] = [0 for i in range(max_idx)]

	### pass 2: calculate some stats of data
	for p in range(len(samples)): # for every data point
		point=samples[p]
		label=labels[p]
		for tuple in point.items(): # for every feature
			f = tuple[0]-1 # feat index
			v = tuple[1] # feat value
			sum_f[f] += v
			sum_l_f[label][f] += v
			sumq_l_f[label][f] += v**2
	### now sum_f, sum_l_f, sumq_l_f are done

	### for each feature, calculate f-score
	eps = 1e-12
	for f in range(min_idx-1, max_idx):
		SB = 0
		for la in p_num.keys():
			SB += (p_num[la] * (sum_l_f[la][f]/p_num[la] - sum_f[f]/data_num)**2 )
		SW = eps
		for la in p_num.keys():
			SW += (sumq_l_f[la][f] - (sum_l_f[la][f]**2)/p_num[la]) 
		F[f+1] = SB / SW
	return F

def findMaxIdx( fn ):
	fp = open(fn); line = fp.readline(); max_idx = 0;
	while line:
		line = line.strip();
		if line[0] == "#":
			line = fp.readline(); continue
		elems = line.split(); 
		for e in elems[1:]:
			points = e.split(":")
			p0 = int( points[0] )
			if p0 > max_idx: 
				max_idx = p0
		line = fp.readline()
	return max_idx


def readDat( filename, cols =[-1] ):
	labels = []; samples = []; max_index = 0; nlines = 0
	max_col = max( cols ); min_col = min( cols )
	if cols == [-1]: 
		print "No feature range supplied, using all..."
		no_range = True
	else:
		print "Using only features specified in cols list..."
		no_range = False
	# load training data
	fp = open(filename)
	line = fp.readline()
	while line:
		# added by untitled, allowing data with comments
		line = line.strip();
		if line[0] == "#":
			line = fp.readline(); continue
		elems = line.split(); sample = {}
		for e in elems[1:]:
			points = e.split(":")
			p0 = int( points[0] )
			if not no_range and p0 > max_col:
				break # assumes attribute indices ordered
			if not no_range and p0 < min_col: continue # too small, save a lookup 
			if p0 in cols or no_range:
				p1 = float( points[1] )
				sample[p0] = p1
				if p0 > max_index:
					max_index = p0
		# Even if there was nothing in defined range, append label / instance to lists
		labels.append(float(elems[0])); samples.append(sample)
		line = fp.readline()
		nlines += 1; 
	fp.close()
	return labels,samples,max_index

### Process args
def procArgs():
	if len(sys.argv) not in [2,4,5]:
		print 'Usage: %s train_file start_col stop_col' % sys.argv[0]
		raise SystemExit

	fn = sys.argv[1]; main_only = False
	if len(sys.argv) == 2: 
		col_start = 0; col_stop = -1 # need to find max_idx
	else: 
		col_start = int( sys.argv[2] )
		col_stop  = int( sys.argv[3] )
	
	if len(sys.argv) == 5:
		main_only = True
	
	if not os.path.isdir( PICKLE_DIR ):
		os.mkdir( PICKLE_DIR )
	
	return fn, col_start, col_stop, main_only

### Launch local subprocess, cmd
def launchProc( cmd ):
	print cmd; rc = popen2.Popen4( cmd )
	return rc

### Call I/O, fsc, pickle
def main( fn, col_start, col_stop ):
	labels,samples,max_idx = readDat( fn, range(col_start,col_stop+1) ) # NB: range is not inclusive for upper val
	score_dict, feat_rank = cal_feat_imp( labels, samples )
	[head, tail] = os.path.split(fn)
	pkl_fn = PICKLE_DIR + tail + '.' + str(col_start) + '.' + str(col_stop) + '.fsc.pkl'
	fh = open( pkl_fn, 'w' ); pickle.dump( score_dict, fh ); fh.close()

### Enter either the job queue or the main branch
def enter():
	[fn, col_start, col_stop, main_only] = procArgs()
	lower = 0; np = 0; cmd = []; active = [-1]

	max_idx = findMaxIdx( fn )
	print 'max idx: ' + str( max_idx )

	if col_stop == -1 or col_stop > max_idx: 
		col_stop = max_idx
		
	if not main_only:
		if col_stop - col_start < COLS:
			lower = col_start; top_of_range = col_stop + 1; step = col_stop
		else:
			lower = col_start; top_of_range = col_stop + COLS; step = COLS
			
		for upper in range(step, top_of_range, step):
			cmd.append( sys.argv[0] + ' ' + sys.argv[1] + ' ' + str(lower) + ' ' + str(upper) + ' child_proc' ); lower = upper + 1
		
		for launch_cmd in cmd:
			rc = launchProc( launch_cmd ); active.append( rc )
			np += 1; 
			if np % NPROCS == 0: # NPROCS at a time, wait for last one...
				rc.wait()
			
		job = active.pop()
		while job != -1:
			# print 'Waiting for a straggling job. Standby...'
			job.wait(); job = active.pop()
			
	else:	
		main( fn, col_start, col_stop )
		exit()
		
	# All partial pickles are now written...	
	# Now gather up the pickles, write final files and clean up.
	ls = os.listdir( PICKLE_DIR )
	this_fsc = {}; fsc_dict = {}
	for pn in ls:
		pn = PICKLE_DIR + pn
		fh = open( pn, 'r' ); this_fsc = pickle.load( fh ); fh.close()
		fsc_dict.update( this_fsc )
		# Clean up pickle
		os.remove( pn )

	# Remove pickle directory
	os.rmdir( PICKLE_DIR )

	# NOTE: from this full dict, some missing feature entries may occur 
	# Enter any missing features, with zero fscore.	
	for i in range( col_start, col_stop ):
		if not fsc_dict.has_key( i ):
			fsc_dict[i] = 0
	del fsc_dict[0] # Remove the 0 entry (libsvm files appear to use 1-based idxing)
	
	# Write the fscore dict out ot a pickle
	[head,tail] = os.path.split( fn )
	#fn_out = tail + '.fscore.pkl'; fh = open( fn_out, 'w' ); pickle.dump( fsc_dict, fh ); fh.close()

	# Sort the score dict
	junk, imp_v = sortFscores( fsc_dict )

	# Write (sorted) f-score list in another file
	f_tuples = fsc_dict.items()
	f_tuples.sort(value_cmpf); fd = open("%s.fscore"%tail, 'w')
	for t in f_tuples:
		fd.write("%d: \t%.6f\n"%t)
	fd.close()


### Program entry point
enter()


